@@ -92,6 +92,7 @@ struct l_dbus_classic {
 	struct l_dbus super;
 	void *auth_command;
 	enum auth_state auth_state;
+	bool skip_hello;
 	struct l_hashmap *match_strings;
 	int *fd_buf;
 	unsigned int num_fds;
@@ -180,7 +181,7 @@ done:
 	if (l_queue_isempty(dbus->message_queue))
 		return false;
 
-	/* Only continue sending messges if the connection is ready */
+	/* Only continue sending messages if the connection is ready */
 	return dbus->is_ready;
 }
 
@@ -412,6 +413,11 @@ static bool auth_write_handler(struct l_io *io, void *user_data)
 	if (classic->auth_state == SETUP_DONE) {
 		struct l_dbus_message *message;
 
+		if (classic->skip_hello) {
+			bus_ready(dbus);
+			return true;
+		}
+
 		l_io_set_read_handler(dbus->io, message_read_handler,
 							dbus, NULL);
 
@@ -1020,7 +1026,7 @@ static const struct l_dbus_ops classic_ops = {
 	.name_acquire = classic_name_acquire,
 };
 
-static struct l_dbus *setup_dbus1(int fd, const char *guid)
+static struct l_dbus *setup_dbus1(int fd, const char *guid, bool skip_hello)
 {
 	static const unsigned char creds = 0x00;
 	char uid[6], hexuid[12], *ptr = hexuid;
@@ -1055,6 +1061,7 @@ static struct l_dbus *setup_dbus1(int fd, const char *guid)
 
 	classic->auth_command = l_strdup_printf("AUTH EXTERNAL %s\r\n", hexuid);
 	classic->auth_state = WAITING_FOR_OK;
+	classic->skip_hello = skip_hello;
 
 	dbus->negotiate_unix_fd = true;
 	dbus->support_unix_fd = false;
@@ -1132,7 +1139,7 @@ static struct l_dbus *setup_unix(char *params)
 		return NULL;
 	}
 
-	return setup_dbus1(fd, guid);
+	return setup_dbus1(fd, guid, false);
 }
 
 static bool setup_tcp_cb(struct l_io *io, void *user_data)
@@ -1302,6 +1309,11 @@ LIB_EXPORT struct l_dbus *l_dbus_new_default(enum l_dbus_bus bus)
 	return setup_address(address);
 }
 
+LIB_EXPORT struct l_dbus *l_dbus_new_private(int fd)
+{
+	return setup_dbus1(fd, NULL, true);
+}
+
 LIB_EXPORT void l_dbus_destroy(struct l_dbus *dbus)
 {
 	if (unlikely(!dbus))
@@ -1683,6 +1695,19 @@ LIB_EXPORT bool l_dbus_object_remove_interface(struct l_dbus *dbus,
 
 LIB_EXPORT void *l_dbus_object_get_data(struct l_dbus *dbus, const char *object,
 					const char *interface)
+{
+	if (unlikely(!dbus))
+		return NULL;
+
+	if (unlikely(!dbus->tree))
+		return NULL;
+
+	return _dbus_object_tree_get_interface_data(dbus->tree, object,
+							interface);
+}
+
+LIB_EXPORT bool l_dbus_object_set_data(struct l_dbus *dbus, const char *object,
+				const char *interface, void *user_data)
 {
 	if (unlikely(!dbus))
 		return false;
@@ -1690,8 +1715,8 @@ LIB_EXPORT void *l_dbus_object_get_data(struct l_dbus *dbus, const char *object,
 	if (unlikely(!dbus->tree))
 		return false;
 
-	return _dbus_object_tree_get_interface_data(dbus->tree, object,
-							interface);
+	return _dbus_object_tree_set_interface_data(dbus->tree, object,
+							interface, user_data);
 }
 
 LIB_EXPORT bool l_dbus_object_manager_enable(struct l_dbus *dbus,
@@ -1800,8 +1825,14 @@ LIB_EXPORT unsigned int l_dbus_add_signal_watch(struct l_dbus *dbus,
 	va_start(args, member);
 
 	rule_len = 0;
-	while (va_arg(args, enum l_dbus_match_type) != L_DBUS_MATCH_NONE)
+	while (true) {
+		type = va_arg(args, enum l_dbus_match_type);
+		if (type == L_DBUS_MATCH_NONE)
+			break;
+
+		va_arg(args, const char *);
 		rule_len++;
+	}
 
 	va_end(args);
 
