@@ -85,7 +85,7 @@ int _vli_cmp(const uint64_t *left, const uint64_t *right, unsigned int ndigits)
 	for (i = ndigits - 1; i >= 0; i--) {
 		if (left[i] > right[i])
 			return 1;
-		else if (left[i] < right[i])
+		if (left[i] < right[i])
 			return -1;
 	}
 
@@ -128,9 +128,8 @@ void _vli_rshift1(uint64_t *vli, unsigned int ndigits)
 }
 
 /* Computes result = left + right, returning carry. Can modify in place. */
-static uint64_t vli_add(uint64_t *result, const uint64_t *left,
-							const uint64_t *right,
-							unsigned int ndigits)
+uint64_t _vli_add(uint64_t *result, const uint64_t *left,
+				const uint64_t *right, unsigned int ndigits)
 {
 	uint64_t carry = 0;
 	unsigned int i;
@@ -150,8 +149,7 @@ static uint64_t vli_add(uint64_t *result, const uint64_t *left,
 
 /* Computes result = left - right, returning borrow. Can modify in place. */
 uint64_t _vli_sub(uint64_t *result, const uint64_t *left,
-							const uint64_t *right,
-							unsigned int ndigits)
+				const uint64_t *right, unsigned int ndigits)
 {
 	uint64_t borrow = 0;
 	unsigned int i;
@@ -290,7 +288,7 @@ void _vli_mod_add(uint64_t *result, const uint64_t *left,
 {
 	uint64_t carry;
 
-	carry = vli_add(result, left, right, ndigits);
+	carry = _vli_add(result, left, right, ndigits);
 
 	/* result > mod (result = mod + remainder), so subtract mod to
 	 * get remainder.
@@ -313,7 +311,82 @@ void _vli_mod_sub(uint64_t *result, const uint64_t *left,
 	 * result + mod (with overflow).
 	 */
 	if (borrow)
-		vli_add(result, result, mod, ndigits);
+		_vli_add(result, result, mod, ndigits);
+}
+
+/* Counts the number of 64-bit "digits" in vli. */
+static unsigned int _vli_num_digits(const uint64_t *vli, unsigned int ndigits)
+{
+	int i;
+
+	/* Search from the end until we find a non-zero digit.
+	 * We do it in reverse because we expect that most digits will
+	 * be nonzero.
+	 */
+	for (i = ndigits - 1; i >= 0 && vli[i] == 0; i--);
+
+	return (i + 1);
+}
+
+/* Counts the number of bits required for vli. */
+static unsigned int _vli_num_bits(const uint64_t *vli, unsigned int ndigits)
+{
+	unsigned int i, num_digits;
+	uint64_t digit;
+
+	num_digits = _vli_num_digits(vli, ndigits);
+	if (num_digits == 0)
+		return 0;
+
+	digit = vli[num_digits - 1];
+	for (i = 0; digit; i++)
+		digit >>= 1;
+
+	return ((num_digits - 1) * 64 + i);
+}
+
+/* Computes result = product % mod, where product is 2N words long.
+ * Currently only designed to work for curve_p or curve_n.
+ */
+void _vli_mmod_slow(uint64_t *result, const uint64_t *product,
+				const uint64_t *mod, unsigned int ndigits)
+{
+	uint64_t mod_m[2 * L_ECC_MAX_DIGITS];
+	uint64_t tmp[2 * L_ECC_MAX_DIGITS];
+	uint64_t *v[2] = { tmp, (uint64_t *) product };
+	uint64_t carry = 0;
+	unsigned int i;
+	/* Shift mod so its highest set bit is at the maximum position. */
+	int shift = (ndigits * 2 * 64) - _vli_num_bits(mod, ndigits);
+	int word_shift = shift / 64;
+	int bit_shift = shift % 64;
+
+	vli_clear(mod_m, word_shift);
+	if (bit_shift > 0) {
+		for (i = 0; i < ndigits; ++i) {
+			mod_m[word_shift + i] = (mod[i] << bit_shift) | carry;
+			carry = mod[i] >> (64 - bit_shift);
+		}
+	} else
+		vli_set(mod_m + word_shift, mod, ndigits);
+
+	for (i = 1; shift >= 0; --shift) {
+		uint64_t borrow = 0;
+		unsigned int j;
+
+		for (j = 0; j < ndigits * 2; ++j) {
+			uint64_t diff = v[i][j] - mod_m[j] - borrow;
+
+			if (diff != v[i][j])
+				borrow = (diff > v[i][j]);
+			v[1 - i][j] = diff;
+		}
+		i = !(i ^ borrow); /* Swap the index if there was no borrow */
+		_vli_rshift1(mod_m, ndigits);
+		mod_m[ndigits - 1] |= mod_m[ndigits] << (64 - 1);
+		_vli_rshift1(mod_m + ndigits, ndigits);
+	}
+	vli_set(result, v[i], ndigits);
 }
 
 /* Computes p_result = p_product % curve_p.
@@ -329,21 +402,70 @@ static void vli_mmod_fast_192(uint64_t *result, const uint64_t *product,
 	vli_set(result, product, ndigits);
 
 	vli_set(tmp, &product[3], ndigits);
-	carry = vli_add(result, result, tmp, ndigits);
+	carry = _vli_add(result, result, tmp, ndigits);
 
 	tmp[0] = 0;
 	tmp[1] = product[3];
 	tmp[2] = product[4];
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	tmp[0] = tmp[1] = product[5];
 	tmp[2] = 0;
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	while (carry || _vli_cmp(curve_prime, result, ndigits) != 1)
 		carry -= _vli_sub(result, result, curve_prime, ndigits);
 }
 
+/* Computes result = product % curve_prime
+ * from http://www.nsa.gov/ia/_files/nist-routines.pdf
+ */
+static void vli_mmod_fast_224(uint64_t *result, const uint64_t *product,
+				const uint64_t *curve_prime, uint64_t *tmp)
+{
+	int carry = 0;
+	const unsigned int ndigits = 4;
+
+	/* t */
+	vli_set(result, product, ndigits);
+	result[3] &= 0xffffffff;
+
+	/* s1 */
+	tmp[0] = 0;
+	tmp[1] = product[3] & 0xffffffff00000000ull;
+	tmp[2] = product[4];
+	tmp[3] = product[5] & 0xffffffff;
+	_vli_add(result, result, tmp, ndigits);
+
+	/* s2 */
+	tmp[1] = product[5] & 0xffffffff00000000ull;
+	tmp[2] = product[6];
+	tmp[3] = 0;
+	_vli_add(result, result, tmp, ndigits);
+
+	/* d1 */
+	tmp[0] = (product[3] >> 32) | (product[4] << 32);
+	tmp[1] = (product[4] >> 32) | (product[5] << 32);
+	tmp[2] = (product[5] >> 32) | (product[6] << 32);
+	tmp[3] = product[6] >> 32;
+	carry -= _vli_sub(result, result, tmp, ndigits);
+
+	/* d2 */
+	tmp[0] = (product[5] >> 32) | (product[6] << 32);
+	tmp[1] = product[6] >> 32;
+	tmp[2] = tmp[3] = 0;
+	carry -= _vli_sub(result, result, tmp, ndigits);
+
+	if (carry < 0) {
+		do {
+			carry += _vli_add(result, result, curve_prime, ndigits);
+		} while (carry < 0);
+	} else {
+		while (carry || _vli_cmp(curve_prime, result, ndigits) != 1)
+			carry -= _vli_sub(result, result, curve_prime, ndigits);
+	}
+}
+
 /* Computes result = product % curve_prime
  * from http://www.nsa.gov/ia/_files/nist-routines.pdf
  */
@@ -362,28 +484,28 @@ static void vli_mmod_fast_256(uint64_t *result, const uint64_t *product,
 	tmp[2] = product[6];
 	tmp[3] = product[7];
 	carry = vli_lshift(tmp, tmp, 1, ndigits);
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s2 */
 	tmp[1] = product[6] << 32;
 	tmp[2] = (product[6] >> 32) | (product[7] << 32);
 	tmp[3] = product[7] >> 32;
 	carry += vli_lshift(tmp, tmp, 1, ndigits);
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s3 */
 	tmp[0] = product[4];
 	tmp[1] = product[5] & 0xffffffff;
 	tmp[2] = 0;
 	tmp[3] = product[7];
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s4 */
 	tmp[0] = (product[4] >> 32) | (product[5] << 32);
 	tmp[1] = (product[5] >> 32) | (product[6] & 0xffffffff00000000ull);
 	tmp[2] = product[7];
 	tmp[3] = (product[6] >> 32) | (product[4] << 32);
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* d1 */
 	tmp[0] = (product[5] >> 32) | (product[6] << 32);
@@ -415,7 +537,7 @@ static void vli_mmod_fast_256(uint64_t *result, const uint64_t *product,
 
 	if (carry < 0) {
 		do {
-			carry += vli_add(result, result, curve_prime, ndigits);
+			carry += _vli_add(result, result, curve_prime, ndigits);
 		} while (carry < 0);
 	} else {
 		while (carry || _vli_cmp(curve_prime, result, ndigits) != 1)
@@ -465,7 +587,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 	tmp[4] = 0;
 	tmp[5] = 0;
 	carry = vli_lshift(tmp, tmp, 1, ndigits);
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s2 */
 	tmp[0] = product[6];
@@ -474,7 +596,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 	tmp[3] = product[9];
 	tmp[4] = product[10];
 	tmp[5] = product[11];
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s3 */
 	tmp[0] = ECC_SET_S(product, 22, 21);
@@ -483,7 +605,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 	tmp[3] = ECC_SET_S(product, 16, 15);
 	tmp[4] = ECC_SET_S(product, 18, 17);
 	tmp[5] = ECC_SET_S(product, 20, 19);
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s4 */
 	tmp[0] = ECC_SET_S(product, 23, -1);
@@ -492,7 +614,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 	tmp[3] = ECC_SET_S(product, 15, 14);
 	tmp[4] = ECC_SET_S(product, 17, 16);
 	tmp[5] = ECC_SET_S(product, 19, 18);
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s5 */
 	tmp[0] = 0;
@@ -501,7 +623,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 	tmp[3] = ECC_SET_S(product, 23, 22);
 	tmp[4] = 0;
 	tmp[5] = 0;
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s6 */
 	tmp[0] = ECC_SET_S(product, -1, 20);
@@ -510,7 +632,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 	tmp[3] = 0;
 	tmp[4] = 0;
 	tmp[5] = 0;
-	carry += vli_add(result, result, tmp, ndigits);
+	carry += _vli_add(result, result, tmp, ndigits);
 
 	/* s7 */
 	tmp[0] = ECC_SET_S(product, 12, 23);
@@ -541,7 +663,7 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 
 	if (carry < 0) {
 		do {
-			carry += vli_add(result, result, curve_prime, ndigits);
+			carry += _vli_add(result, result, curve_prime, ndigits);
 		} while (carry < 0);
 	} else {
 		while (carry || _vli_cmp(curve_prime, result, ndigits) != 1)
@@ -551,23 +673,49 @@ static void vli_mmod_fast_384(uint64_t *result, const uint64_t *product,
 
 /* Computes result = product % curve_prime
  *  from http://www.nsa.gov/ia/_files/nist-routines.pdf
-*/
-static bool vli_mmod_fast(uint64_t *result, uint64_t *product,
-				const uint64_t *curve_prime,
-				unsigned int ndigits)
+ */
+static void vli_mmod_fast_521(uint64_t *result, const uint64_t *product,
+				const uint64_t *curve_prime, uint64_t *tmp)
+{
+	const unsigned int ndigits = 9;
+	size_t i;
+
+	/* Initialize result with lowest 521 bits from product */
+	vli_set(result, product, ndigits);
+	result[8] &= 0x1ff;
+
+	for (i = 0; i < ndigits; i++)
+		tmp[i] = (product[8 + i] >> 9) | (product[9 + i] << 55);
+	tmp[8] &= 0x1ff;
+
+	_vli_mod_add(result, result, tmp, curve_prime, ndigits);
+}
+
+/* Computes result = product % curve_prime
+ *  from http://www.nsa.gov/ia/_files/nist-routines.pdf
+ */
+bool _vli_mmod_fast(uint64_t *result, const uint64_t *product,
+			const uint64_t *curve_prime, unsigned int ndigits)
 {
 	uint64_t tmp[2 * L_ECC_MAX_DIGITS];
+	unsigned int nbits = _vli_num_bits(curve_prime, ndigits);
 
-	switch (ndigits) {
-	case 3:
+	switch (nbits) {
+	case 192:
 		vli_mmod_fast_192(result, product, curve_prime, tmp);
 		break;
-	case 4:
+	case 224:
+		vli_mmod_fast_224(result, product, curve_prime, tmp);
+		break;
+	case 256:
 		vli_mmod_fast_256(result, product, curve_prime, tmp);
 		break;
-	case 6:
+	case 384:
 		vli_mmod_fast_384(result, product, curve_prime, tmp);
 		break;
+	case 521:
+		vli_mmod_fast_521(result, product, curve_prime, tmp);
+		break;
 	default:
 		return false;
 	}
@@ -583,7 +731,7 @@ void _vli_mod_mult_fast(uint64_t *result, const uint64_t *left,
 	uint64_t product[2 * L_ECC_MAX_DIGITS];
 
 	vli_mult(product, left, right, ndigits);
-	vli_mmod_fast(result, product, curve_prime, ndigits);
+	_vli_mmod_fast(result, product, curve_prime, ndigits);
 }
 
 /* Computes result = left^2 % curve_p. */
@@ -594,7 +742,7 @@ void _vli_mod_square_fast(uint64_t *result, const uint64_t *left,
 	uint64_t product[2 * L_ECC_MAX_DIGITS];
 
 	vli_square(product, left, ndigits);
-	vli_mmod_fast(result, product, curve_prime, ndigits);
+	_vli_mmod_fast(result, product, curve_prime, ndigits);
 }
 
 #define EVEN(vli) (!(vli[0] & 1))
@@ -629,7 +777,7 @@ void _vli_mod_inv(uint64_t *result, const uint64_t *input,
 			_vli_rshift1(a, ndigits);
 
 			if (!EVEN(u))
-				carry = vli_add(u, u, mod, ndigits);
+				carry = _vli_add(u, u, mod, ndigits);
 
 			_vli_rshift1(u, ndigits);
 			if (carry)
@@ -638,7 +786,7 @@ void _vli_mod_inv(uint64_t *result, const uint64_t *input,
 			_vli_rshift1(b, ndigits);
 
 			if (!EVEN(v))
-				carry = vli_add(v, v, mod, ndigits);
+				carry = _vli_add(v, v, mod, ndigits);
 
 			_vli_rshift1(v, ndigits);
 			if (carry)
@@ -648,11 +796,11 @@ void _vli_mod_inv(uint64_t *result, const uint64_t *input,
 			_vli_rshift1(a, ndigits);
 
 			if (_vli_cmp(u, v, ndigits) < 0)
-				vli_add(u, u, mod, ndigits);
+				_vli_add(u, u, mod, ndigits);
 
 			_vli_sub(u, u, v, ndigits);
 			if (!EVEN(u))
-				carry = vli_add(u, u, mod, ndigits);
+				carry = _vli_add(u, u, mod, ndigits);
 
 			_vli_rshift1(u, ndigits);
 			if (carry)
@@ -662,11 +810,11 @@ void _vli_mod_inv(uint64_t *result, const uint64_t *input,
 			_vli_rshift1(b, ndigits);
 
 			if (_vli_cmp(v, u, ndigits) < 0)
-				vli_add(v, v, mod, ndigits);
+				_vli_add(v, v, mod, ndigits);
 
 			_vli_sub(v, v, u, ndigits);
 			if (!EVEN(v))
-				carry = vli_add(v, v, mod, ndigits);
+				carry = _vli_add(v, v, mod, ndigits);
 
 			_vli_rshift1(v, ndigits);
 			if (carry)
@@ -720,7 +868,7 @@ static void ecc_point_double_jacobian(uint64_t *x1, uint64_t *y1, uint64_t *z1,
 	/* t1 = 3*(x1^2 - z1^4) */
 	_vli_mod_add(x1, x1, z1, curve_prime, ndigits);
 	if (vli_test_bit(x1, 0)) {
-		uint64_t carry = vli_add(x1, x1, curve_prime, ndigits);
+		uint64_t carry = _vli_add(x1, x1, curve_prime, ndigits);
 		_vli_rshift1(x1, ndigits);
 		x1[ndigits - 1] |= carry << 63;
 	} else {
@@ -892,31 +1040,36 @@ void _ecc_point_mult(struct l_ecc_point *result,
 	uint64_t sk[2][L_ECC_MAX_DIGITS];
 	int i, nb;
 	unsigned int ndigits = curve->ndigits;
+	unsigned int nbits = _vli_num_bits(curve->n, ndigits);
 	int num_bits;
 	int carry;
 
-	carry = vli_add(sk[0], scalar, curve->n, ndigits);
-	vli_add(sk[1], sk[0], curve->n, ndigits);
+	carry = _vli_add(sk[0], scalar, curve->n, ndigits);
+	_vli_add(sk[1], sk[0], curve->n, ndigits);
 	scalar = sk[!carry];
-	num_bits = sizeof(uint64_t) * ndigits * 8 + 1;
+	/* secp224r1 and secp521r1 curves */
+	if (nbits == 224 || nbits == 521)
+		num_bits = nbits + 2;
+	else
+		num_bits = sizeof(uint64_t) * ndigits * 8 + 1;
 
 	vli_set(rx[1], point->x, ndigits);
 	vli_set(ry[1], point->y, ndigits);
 
 	xycz_initial_double(rx[1], ry[1], rx[0], ry[0], initial_z, curve_prime,
-			    ndigits);
+				ndigits);
 
 	for (i = num_bits - 2; i > 0; i--) {
 		nb = !vli_test_bit(scalar, i);
 		xycz_add_c(rx[1 - nb], ry[1 - nb], rx[nb], ry[nb], curve_prime,
-			   ndigits);
+				ndigits);
 		xycz_add(rx[nb], ry[nb], rx[1 - nb], ry[1 - nb], curve_prime,
-			 ndigits);
+				ndigits);
 	}
 
 	nb = !vli_test_bit(scalar, 0);
 	xycz_add_c(rx[1 - nb], ry[1 - nb], rx[nb], ry[nb], curve_prime,
-		   ndigits);
+			ndigits);
 
 	/* Find final 1/Z value. */
 	/* X1 - X0 */
