@@ -1,49 +1,54 @@
 /*
+ * Embedded Linux library
+ * Copyright (C) 2011-2014  Intel Corporation
  *
- *  Embedded Linux library
- *
- *  Copyright (C) 2011-2014  Intel Corporation. All rights reserved.
- *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU Lesser General Public
- *  License as published by the Free Software Foundation; either
- *  version 2.1 of the License, or (at your option) any later version.
- *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  Lesser General Public License for more details.
- *
- *  You should have received a copy of the GNU Lesser General Public
- *  License along with this library; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
- *
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
+#if __STDC_VERSION__ <= 199409L
+#define _DEFAULT_SOURCE  /* for strto{u}ll() */
+#endif
+
+#define _GNU_SOURCE
 #include <stdio.h>
-#include <stdlib.h>
 #include <errno.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <stdlib.h>
+#include <string.h>
+#include <strings.h>
+#include <stdbool.h>
+#include <stdint.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/mman.h>
 
-#include "util.h"
+#include "useful.h"
+#include "strv.h"
+#include "utf8.h"
 #include "string.h"
 #include "queue.h"
 #include "settings.h"
 #include "private.h"
+#include "missing.h"
+#include "pem-private.h"
 
 struct setting_data {
 	char *key;
 	char *value;
 };
 
+struct embedded_group_data {
+	char *name;
+	char type[32];
+	size_t len;
+	char data[];
+};
+
 struct group_data {
 	char *name;
 	struct l_queue *settings;
@@ -54,6 +59,7 @@ struct l_settings {
 	l_settings_destroy_cb_t debug_destroy;
 	void *debug_data;
 	struct l_queue *groups;
+	struct l_queue *embedded_groups;
 };
 
 static void setting_destroy(void *data)
@@ -61,6 +67,7 @@ static void setting_destroy(void *data)
 	struct setting_data *pair = data;
 
 	l_free(pair->key);
+	explicit_bzero(pair->value, strlen(pair->value));
 	l_free(pair->value);
 	l_free(pair);
 }
@@ -75,16 +82,82 @@ static void group_destroy(void *data)
 	l_free(group);
 }
 
+static void embedded_group_destroy(void *data)
+{
+	struct embedded_group_data *group = data;
+
+	l_free(group->name);
+	l_free(group);
+}
+
 LIB_EXPORT struct l_settings *l_settings_new(void)
 {
 	struct l_settings *settings;
 
 	settings = l_new(struct l_settings, 1);
 	settings->groups = l_queue_new();
+	settings->embedded_groups = l_queue_new();
 
 	return settings;
 }
 
+static void copy_key_value_foreach(void *data, void *user_data)
+{
+	struct setting_data *s = data;
+	struct l_queue *settings = user_data;
+	struct setting_data *copy = l_new(struct setting_data, 1);
+
+	copy->key = l_strdup(s->key);
+	copy->value = l_strdup(s->value);
+
+	l_queue_push_head(settings, copy);
+}
+
+static void copy_group_foreach(void *data, void *user_data)
+{
+	struct group_data *group = data;
+	struct l_queue *groups = user_data;
+	struct group_data *copy = l_new(struct group_data, 1);
+
+	copy->name = l_strdup(group->name);
+	copy->settings = l_queue_new();
+
+	l_queue_push_head(groups, copy);
+
+	l_queue_foreach(group->settings, copy_key_value_foreach,
+				copy->settings);
+}
+
+static void copy_embedded_foreach(void *data, void *user_data)
+{
+	struct embedded_group_data *embedded = data;
+	struct l_queue *groups = user_data;
+	struct embedded_group_data *copy;
+
+	copy = l_memdup(embedded, sizeof(struct embedded_group_data) +
+			embedded->len + 1);
+	copy->name = l_strdup(embedded->name);
+
+	l_queue_push_tail(groups, copy);
+}
+
+LIB_EXPORT struct l_settings *l_settings_clone(
+					const struct l_settings *settings)
+{
+	struct l_settings *copy;
+
+	if (unlikely(!settings))
+		return NULL;
+
+	copy = l_settings_new();
+
+	l_queue_foreach(settings->groups, copy_group_foreach, copy->groups);
+	l_queue_foreach(settings->embedded_groups, copy_embedded_foreach,
+				copy->embedded_groups);
+
+	return copy;
+}
+
 LIB_EXPORT void l_settings_free(struct l_settings *settings)
 {
 	if (unlikely(!settings))
@@ -94,6 +167,7 @@ LIB_EXPORT void l_settings_free(struct l_settings *settings)
 		settings->debug_destroy(settings->debug_data);
 
 	l_queue_destroy(settings->groups, group_destroy);
+	l_queue_destroy(settings->embedded_groups, embedded_group_destroy);
 
 	l_free(settings);
 }
@@ -131,6 +205,7 @@ static char *unescape_value(const char *value)
 			*n = '\\';
 			break;
 		default:
+			explicit_bzero(ret, n - ret);
 			l_free(ret);
 			return NULL;
 		}
@@ -158,7 +233,7 @@ static char *escape_value(const char *value)
 		case '\r':
 		case '\\':
 			j += 1;
-			/* Fall through */
+			/* fall through */
 		default:
 			lead_whitespace = false;
 		}
@@ -198,6 +273,7 @@ static char *escape_value(const char *value)
 			ret[j++] = '\\';
 			ret[j++] = '\\';
 			lead_whitespace = false;
+			break;
 		default:
 			ret[j++] = value[i];
 			lead_whitespace = false;
@@ -209,6 +285,140 @@ static char *escape_value(const char *value)
 	return ret;
 }
 
+static ssize_t parse_pem(const char *data, size_t len)
+{
+	const char *ptr;
+	const char *end;
+	size_t count = 0;
+
+	ptr = data;
+	end = data + len;
+
+	while (ptr && ptr < end) {
+		const char *pem_start = ptr;
+
+		if (!pem_next(ptr, len, NULL, NULL, &ptr, true)) {
+			if (ptr)
+				return -EINVAL;
+
+			break;
+		}
+
+		len -= ptr - pem_start;
+		count += ptr - pem_start;
+	}
+
+	return count;
+}
+
+struct group_extension {
+	char *name;
+	ssize_t (*parse)(const char *data, size_t len);
+};
+
+static const struct group_extension pem_extension = {
+	.name = "pem",
+	.parse = parse_pem,
+};
+
+static const struct group_extension *extensions[] = {
+	&pem_extension,
+	NULL
+};
+
+static const struct group_extension *find_group_extension(const char *type,
+								size_t len)
+{
+	unsigned int i;
+
+	for (i = 0; extensions[i]; i++) {
+		if (!strncmp(type, extensions[i]->name, len))
+			return extensions[i];
+	}
+
+	return NULL;
+}
+
+static ssize_t parse_embedded_group(struct l_settings *setting,
+					const char *data,
+					size_t line_len, size_t len,
+					size_t line)
+{
+	struct embedded_group_data *group;
+	const struct group_extension *ext;
+	const char *ptr;
+	const char *type;
+	size_t type_len;
+	const char *name;
+	size_t name_len;
+	ssize_t bytes;
+
+	/* Must be at least [@a@b] */
+	if (line_len < 6)
+		goto invalid_group;
+
+	/* caller checked data[1] == '@', next char is type */
+	type = data + 2;
+
+	ptr = memchr(type, '@', line_len - 2);
+
+	type_len = ptr - type;
+
+	if (!ptr || type_len > 31 || type_len < 1)
+		goto invalid_group;
+
+	if (ptr + 1 > data + line_len)
+		goto invalid_group;
+
+	name = ptr + 1;
+
+	/* subtract [@@ + type */
+	ptr = memchr(name, ']', line_len - 3 - type_len);
+
+	name_len = ptr - name;
+
+	if (!ptr || name_len < 1)
+		goto invalid_group;
+
+	ext = find_group_extension(type, type_len);
+	if (!ext)
+		goto invalid_group;
+
+	if (ptr + 2 > data + len) {
+		l_util_debug(setting->debug_handler, setting->debug_data,
+				"Embedded group had no payload");
+		return -EINVAL;
+	}
+
+	bytes = ext->parse(ptr + 2, len - line_len);
+	if (bytes < 0) {
+		l_util_debug(setting->debug_handler, setting->debug_data,
+				"Failed to parse embedded group data");
+		return -EINVAL;
+	}
+
+	group = l_malloc(sizeof(struct embedded_group_data) + bytes + 1);
+
+	group->name = l_strndup(name, name_len);
+
+	memcpy(group->type, type, type_len);
+	group->type[type_len] = '\0';
+
+	group->len = bytes;
+	memcpy(group->data, ptr + 2, bytes);
+	group->data[bytes] = '\0';
+
+	l_queue_push_tail(setting->embedded_groups, group);
+
+	return bytes;
+
+invalid_group:
+	l_util_debug(setting->debug_handler, setting->debug_data,
+			"Invalid embedded group at line %zd", line);
+
+	return -EINVAL;
+}
+
 static bool parse_group(struct l_settings *settings, const char *data,
 			size_t len, size_t line)
 {
@@ -254,6 +464,17 @@ static bool parse_group(struct l_settings *settings, const char *data,
 	return true;
 }
 
+static bool validate_key_character(char c)
+{
+	if (l_ascii_isalnum(c))
+		return true;
+
+	if (c == '_' || c == '-' || c == '.')
+		return true;
+
+	return false;
+}
+
 static unsigned int parse_key(struct l_settings *settings, const char *data,
 				size_t len, size_t line)
 {
@@ -263,10 +484,7 @@ static unsigned int parse_key(struct l_settings *settings, const char *data,
 	struct setting_data *pair;
 
 	for (i = 0; i < len; i++) {
-		if (l_ascii_isalnum(data[i]))
-			continue;
-
-		if (data[i] == '_' || data[i] == '-')
+		if (validate_key_character(data[i]))
 			continue;
 
 		if (l_ascii_isblank(data[i]))
@@ -342,7 +560,7 @@ static bool parse_keyvalue(struct l_settings *settings, const char *data,
 		return false;
 	}
 
-	if (parse_key(settings, data, equal - data, line) == false)
+	if (!parse_key(settings, data, equal - data, line))
 		return false;
 
 	equal += 1;
@@ -357,6 +575,7 @@ LIB_EXPORT bool l_settings_load_from_data(struct l_settings *settings,
 {
 	size_t pos = 0;
 	bool r = true;
+	bool has_group = false;
 	const char *eol;
 	size_t line = 1;
 	size_t line_len;
@@ -382,11 +601,31 @@ LIB_EXPORT bool l_settings_load_from_data(struct l_settings *settings,
 
 		line_len = eol - data - pos;
 
-		if (data[pos] == '[')
+		if (line_len > 1 && data[pos] == '[' && data[pos + 1] == '@') {
+			ssize_t ret;
+
+			ret = parse_embedded_group(settings, data + pos,
+							line_len, len - pos,
+							line);
+			if (ret < 0)
+				return false;
+
+			/*
+			 * This is the offset for the actual raw data, the
+			 * group line will be offset below
+			 */
+			pos += ret;
+		} else if (data[pos] == '[') {
 			r = parse_group(settings, data + pos, line_len, line);
-		else if (data[pos] != '#')
+			if (r)
+				has_group = true;
+		} else if (data[pos] != '#') {
+			if (!has_group)
+				return false;
+
 			r = parse_keyvalue(settings, data + pos, line_len,
 						line);
+		}
 
 		pos += line_len;
 	}
@@ -409,11 +648,12 @@ LIB_EXPORT char *l_settings_to_data(const struct l_settings *settings,
 	group_entry = l_queue_get_entries(settings->groups);
 	while (group_entry) {
 		struct group_data *group = group_entry->data;
-		const struct l_queue_entry *setting_entry =
-				l_queue_get_entries(group->settings);
+		const struct l_queue_entry *setting_entry;
 
 		l_string_append_printf(buf, "[%s]\n", group->name);
 
+		setting_entry = l_queue_get_entries(group->settings);
+
 		while (setting_entry) {
 			struct setting_data *setting = setting_entry->data;
 
@@ -428,6 +668,24 @@ LIB_EXPORT char *l_settings_to_data(const struct l_settings *settings,
 		group_entry = group_entry->next;
 	}
 
+	group_entry = l_queue_get_entries(settings->embedded_groups);
+
+	if (group_entry && l_queue_length(settings->groups) > 0)
+		l_string_append_c(buf, '\n');
+
+	while (group_entry) {
+		struct embedded_group_data *group = group_entry->data;
+
+		l_string_append_printf(buf, "[@%s@%s]\n%s",
+					group->type,
+					group->name,
+					group->data);
+		if (group_entry->next)
+			l_string_append_c(buf, '\n');
+
+		group_entry = group_entry->next;
+	}
+
 	ret = l_string_unwrap(buf);
 
 	if (len)
@@ -464,6 +722,12 @@ LIB_EXPORT bool l_settings_load_from_file(struct l_settings *settings,
 		return false;
 	}
 
+	/* Nothing to do, assume success */
+	if (st.st_size == 0) {
+		close(fd);
+		return true;
+	}
+
 	data = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
 	if (data == MAP_FAILED) {
 		l_util_debug(settings->debug_handler, settings->debug_data,
@@ -644,18 +908,41 @@ static bool validate_group_name(const char *group_name)
 	return true;
 }
 
+LIB_EXPORT bool l_settings_add_group(struct l_settings *settings,
+					const char *group_name)
+{
+	struct group_data *group;
+
+	if (unlikely(!settings || !group_name))
+		return false;
+
+	if (!validate_group_name(group_name)) {
+		l_util_debug(settings->debug_handler, settings->debug_data,
+				"Invalid group name %s", group_name);
+		return false;
+	}
+
+	group = l_queue_find(settings->groups, group_match, group_name);
+	if (group) {
+		l_util_debug(settings->debug_handler, settings->debug_data,
+				"Group %s exists", group_name);
+		return true;
+	}
+
+	group = l_new(struct group_data, 1);
+	group->name = l_strdup(group_name);
+	group->settings = l_queue_new();
+	l_queue_push_tail(settings->groups, group);
+	return true;
+}
+
 static bool validate_key(const char *key)
 {
 	int i;
 
 	for (i = 0; key[i]; i++) {
-		if (l_ascii_isalnum(key[i]))
-			continue;
-
-		if (key[i] == '_' || key[i] == '-')
-			continue;
-
-		return false;
+		if (!validate_key_character(key[i]))
+			return false;
 	}
 
 	return true;
@@ -670,17 +957,13 @@ static bool set_value(struct l_settings *settings, const char *group_name,
 	if (!validate_group_name(group_name)) {
 		l_util_debug(settings->debug_handler, settings->debug_data,
 				"Invalid group name %s", group_name);
-		l_free(value);
-
-		return false;
+		goto error;
 	}
 
 	if (!validate_key(key)) {
 		l_util_debug(settings->debug_handler, settings->debug_data,
 				"Invalid key %s", key);
-		l_free(value);
-
-		return false;
+		goto error;
 	}
 
 	group = l_queue_find(settings->groups, group_match, group_name);
@@ -704,10 +987,16 @@ add_pair:
 		return true;
 	}
 
+	explicit_bzero(pair->value, strlen(pair->value));
 	l_free(pair->value);
 	pair->value = value;
 
 	return true;
+
+error:
+	explicit_bzero(value, strlen(value));
+	l_free(value);
+	return false;
 }
 
 LIB_EXPORT bool l_settings_set_value(struct l_settings *settings,
@@ -783,7 +1072,7 @@ LIB_EXPORT bool l_settings_get_int(const struct l_settings *settings,
 
 	errno = 0;
 
-	t = r = strtol(value, &endp, 10);
+	t = r = strtol(value, &endp, 0);
 	if (*endp != '\0')
 		goto error;
 
@@ -814,32 +1103,18 @@ LIB_EXPORT bool l_settings_set_int(struct l_settings *settings,
 }
 
 LIB_EXPORT bool l_settings_get_uint(const struct l_settings *settings,
-					const char *group_name, const char *key,
+					const char *group_name,
+					const char *key,
 					unsigned int *out)
 {
 	const char *value = l_settings_get_value(settings, group_name, key);
-	unsigned long int r;
-	unsigned int t;
-	char *endp;
 
 	if (!value)
 		return false;
 
-	if (*value == '\0')
+	if (l_safe_atou32(value, out) < 0)
 		goto error;
 
-	errno = 0;
-
-	t = r = strtoul(value, &endp, 10);
-	if (*endp != '\0')
-		goto error;
-
-	if (unlikely(errno == ERANGE || r != t))
-		goto error;
-
-	if (out)
-		*out = r;
-
 	return true;
 
 error:
@@ -876,7 +1151,7 @@ LIB_EXPORT bool l_settings_get_int64(const struct l_settings *settings,
 
 	errno = 0;
 
-	r = strtoll(value, &endp, 10);
+	r = strtoll(value, &endp, 0);
 	if (*endp != '\0')
 		goto error;
 
@@ -907,7 +1182,8 @@ LIB_EXPORT bool l_settings_set_int64(struct l_settings *settings,
 }
 
 LIB_EXPORT bool l_settings_get_uint64(const struct l_settings *settings,
-					const char *group_name, const char *key,
+					const char *group_name,
+					const char *key,
 					uint64_t *out)
 {
 	const char *value = l_settings_get_value(settings, group_name, key);
@@ -917,12 +1193,13 @@ LIB_EXPORT bool l_settings_get_uint64(const struct l_settings *settings,
 	if (!value)
 		return false;
 
-	if (*value == '\0')
+	/* Do not allow '+' or '-' or empty string */
+	if (!l_ascii_isdigit(*value))
 		goto error;
 
 	errno = 0;
 
-	r = strtoull(value, &endp, 10);
+	r = strtoull(value, &endp, 0);
 	if (*endp != '\0')
 		goto error;
 
@@ -1057,7 +1334,7 @@ LIB_EXPORT bool l_settings_set_double(struct l_settings *settings,
 {
 	L_AUTO_FREE_VAR(char *, buf);
 
-	buf = l_strdup_printf("%d", in);
+	buf = l_strdup_printf("%f", in);
 
 	return l_settings_set_value(settings, group_name, key, buf);
 }
@@ -1103,11 +1380,48 @@ LIB_EXPORT bool l_settings_set_float(struct l_settings *settings,
 {
 	L_AUTO_FREE_VAR(char *, buf);
 
-	buf = l_strdup_printf("%f", in);
+	buf = l_strdup_printf("%f", (double)in);
 
 	return l_settings_set_value(settings, group_name, key, buf);
 }
 
+LIB_EXPORT uint8_t *l_settings_get_bytes(const struct l_settings *settings,
+						const char *group_name,
+						const char *key,
+						size_t *out_len)
+{
+	const char *value = l_settings_get_value(settings, group_name, key);
+
+	if (!value)
+		return NULL;
+
+	if (value[0] == '\0') {
+		*out_len = 0;
+
+		/* Return something that can be l_freed but is not a NULL */
+		return l_memdup("", 1);
+	}
+
+	return l_util_from_hexstring(value, out_len);
+}
+
+LIB_EXPORT bool l_settings_set_bytes(struct l_settings *settings,
+					const char *group_name, const char *key,
+					const uint8_t *value, size_t value_len)
+{
+	char *buf;
+
+	if (unlikely(!settings || !value))
+		return false;
+
+	if (value_len)
+		buf = l_util_hexstring(value, value_len);
+	else
+		buf = l_strdup("");
+
+	return set_value(settings, group_name, key, buf);
+}
+
 LIB_EXPORT bool l_settings_remove_group(struct l_settings *settings,
 					const char *group_name)
 {
@@ -1147,3 +1461,82 @@ LIB_EXPORT bool l_settings_remove_key(struct l_settings *settings,
 
 	return true;
 }
+
+static void gather_embedded_groups(void *data, void *user_data)
+{
+	struct embedded_group_data *group_data = data;
+	struct gather_data *gather = user_data;
+
+	gather->v[gather->cur++] = l_strdup(group_data->name);
+}
+
+LIB_EXPORT char **l_settings_get_embedded_groups(struct l_settings *settings)
+{
+	char **ret;
+	struct gather_data gather;
+
+	if (unlikely(!settings))
+		return NULL;
+
+	ret = l_new(char *, l_queue_length(settings->groups) + 1);
+	gather.v = ret;
+	gather.cur = 0;
+
+	l_queue_foreach(settings->embedded_groups, gather_embedded_groups,
+				&gather);
+
+	return ret;
+}
+
+static bool embedded_group_match(const void *a, const void *b)
+{
+	const struct embedded_group_data *group = a;
+	const char *name = b;
+
+	return !strcmp(group->name, name);
+}
+
+LIB_EXPORT bool l_settings_has_embedded_group(struct l_settings *settings,
+						const char *group)
+{
+	struct embedded_group_data *group_data;
+
+	if (unlikely(!settings))
+		return false;
+
+	group_data = l_queue_find(settings->embedded_groups,
+					embedded_group_match, group);
+
+	return group_data != NULL;
+}
+
+LIB_EXPORT const char *l_settings_get_embedded_value(
+						struct l_settings *settings,
+						const char *group_name,
+						const char **out_type)
+{
+	struct embedded_group_data *group;
+
+	if (unlikely(!settings))
+		return NULL;
+
+	group = l_queue_find(settings->embedded_groups,
+					embedded_group_match, group_name);
+	if (!group)
+		return NULL;
+
+	if (out_type)
+		*out_type = group->type;
+
+	return group->data;
+}
+
+LIB_EXPORT bool l_settings_remove_embedded_groups(struct l_settings *settings)
+{
+	if (unlikely(!settings))
+		return false;
+
+	l_queue_clear(settings->embedded_groups, embedded_group_destroy);
+
+	return true;
+}
