@@ -202,11 +202,11 @@ static void test_priv_key_from_data(const void *data)
 						strlen(raw_private_key),
 						"abc", &is_encrypted);
 	if (!key) {
-		l_info("* Some kernel versions do not automatically load\n"
-			"* the pkcs8_key_parser module. If the system running\n"
-			"* test has not loaded this module, a failure here is\n"
-			"* likely. Running \"modprobe pkcs8_key_parser\" may\n"
-			"* correct this issue.\n");
+		l_info("# Some kernel versions do not automatically load\n"
+			"# the pkcs8_key_parser module. If the system running\n"
+			"# test has not loaded this module, a failure here is\n"
+			"# likely. Running \"modprobe pkcs8_key_parser\" may\n"
+			"# correct this issue.\n");
 		exit(1);
 	}
 
